
from .base import *
from ... import dz,pyz
from ..ioc.single import Single
class ObjectEncape(BaseEncape):
    '''
        id=?
        
    '''
    def init(self, single, src, args, maps, sets, before_set=None, after_set=None):
        super().init()
        self.single = Single(single)
        self.src, self.args, self.maps, self.sets = src, args, maps, sets
        self.before_set = before_set
        self.after_set = after_set
        self.objs = {}
    def call(self, params=None, **maps):
        obj, find = self.single.get(params)
        if find:
            return obj
        src = self.src
        if params is None:
            obj_conf = Params(obj = self)
        else:
            obj_conf = params.clone(obj=self)
        if isinstance(src, Encape):
            src = src(obj_conf)
        args = [self.obj(k,obj_conf) for k in self.args]
        _maps = {self.obj(k,obj_conf):self.obj(v,obj_conf) for k,v in self.maps}
        obj = src(*args, **_maps)
        self.obj(self.before_set, obj_conf)
        for k,v in self.sets:
            setattr(obj, self.obj(k,obj_conf), self.obj(v,obj_conf))
        self.obj(self.after_set, obj_conf)
        self.single.set(params, obj)
        return obj
class ObjectDeal(BaseDeal):
    def init(self):
        super().init()
        self.load_srcs = {}
    def build(self, conf, unit):
        id,id_find = unit.conf_key(conf)
        single = dz.g(conf, single=None)
        if single is None and not id_find:
            single = Single.Key.multi
        src, args, maps, sets = dz.g(conf, source=None, args=[], maps={},sets=[])
        before_set, after_set = dz.g(conf, before_set=None, after_set=None)
        before_set = self.get_encape(before_set, unit)
        after_set = self.get_encape(after_set, unit)
        if Confs.is_conf(src):
            src = unit.get_encape(src, unit)
        elif type(src)==str:
            if src not in self.load_srcs:
                self.load_srcs[src] = pyz.load(src)
            src = self.load_srcs[src]
        args = [self.get_encape(k, unit) for k in args]
        lmaps = []
        for k,v in dz.dict2iter(maps):
            lmaps.append((self.get_encape(k, unit),self.get_encape(v, unit)))
        lsets = []
        for k,v in dz.dict2iter(sets):
            lsets.append((self.get_encape(k, unit),self.get_encape(v, unit)))
        encape = ObjectEncape(single, src, args, lmaps, lsets,before_set,after_set)
        return encape

pass


#

from buildz.iocz.conf.mg import ConfManager

from buildz.iocz.ioc import *
from buildz.iocz.ioc_deal import obj, deal
from buildz import xf, pyz, Base

class Test(Base):
    def init(self, id=0):
        self.id = id

pass
confs = r"""
confs.pri: {
    deal_obj:{
        type=deal
        source=buildz.iocz.ioc_deal.obj.ObjectDeal
        deals: obj
        call=1
    }
}
confs: [
    {
        id=test
        type=obj
        source=buildz.iocz.ioc.test.Test
        args=[
            {
                type=obj
                source = buildz.iocz.ioc.test.Test
                maps={id=123}
            }
        ]
    }
]
builds: [deal_obj]
"""
def test():
    mg = ConfManager()
    mg.set_deal("deal", deal.DealDeal())
    print(mg)
    unit = mg.add_conf(confs)
    encape, tag, find = mg.get_encape("test")
    print(f"encape: {encape}")
    it = encape()
    print(f"obj: {id(it), id(it.id),it.id.id}")
    it = encape()
    print(f"obj: {id(it), id(it.id),it.id.id}")

pyz.lc(locals(), test)


#
from .base import *
from .datas import Datas
from .confs import Confs
from .encapes import Encapes
from ... import pyz
from .builds import Builds
class Unit(Base):
    def init(self, ns=None, deal_ns=None, deal_key = None, conf_key = None, id=None):
        self.ns = ns
        self.deal_ns = deal_ns
        self.deal_key = deal_key
        self.conf_key = conf_key
        self.id = id
        self.confs = Confs(ns, deal_ns, id)
        self.deals = Datas(deal_ns, id)
        self.builds = Builds(self)
        self.mg = None
        self.build_encapes()
    def update_ns(self, ns):
        self.ns = ns
        self.confs.ns = ns
        if self.encapes is not None:
            self.encapes.ns = ns
    def update_deal_ns(self, deal_ns):
        self.deal_ns = deal_ns
        self.deals.ns = deal_ns
    def add_build(self, conf):
        self.builds.add(conf)
    def build(self):
        self.builds.build()
    def is_conf(self, obj):
        return self.confs.is_conf(obj)
    def build_encapes(self):
        if self.confs is None or self.deals is None or self.deal_key is None:
            self.encapes = None
            return
        self.encapes = Encapes(self.ns, self.id, None, self)
    def bind(self, mg):
        if self.mg == mg:
            return
        self.mg = mg
        self.deal_key = pyz.nnull(self.deal_key, mg.deal_key)
        self.conf_key = pyz.nnull(self.conf_key, mg.conf_key)
        if self.id is None:
            self.id = mg.id()
            self.confs.set_id(self.id)
            self.deals.set_id(self.id)
        self.mg.add(self)
        self.builds.bind(mg.builds)
        self.confs.bind(mg.confs)
        self.deals.bind(mg.deals)
        self.build_encapes()
        self.encapes.bind(mg.encapes)
    def get_deal(self, key, src=None, id=None, gfind=True):
        return self.deals.tget(key, src, id, gfind)
    def set_deal(self, key, deal, tag=None):
        self.deals.set(key, deal, tag)
    def get_conf(self, key, src=None, id=None, gfind=True):
        return self.confs.tget(key, src, id, gfind)
    def set_conf(self, key, conf, tag=None):
        self.conf_key.fill(conf, key)
        self.confs.set(key, conf, tag)
    def get_encape(self, key, src=None, id=None, gfind=True):
        self.build()
        return self.encapes.tget(key, src, id, gfind)
    def set_encape(self, key, encape, tag=None):
        self.encapes.set(key, encape, tag)


from ..ioc.base import *
from ..ioc.confs import Confs
from ... import dz
class BaseEncape(Encape):
    @staticmethod
    def obj(val,*a,**b):
        if not isinstance(val, Encape):
            return val
        return val(*a,**b)

pass
class BaseDeal(Deal):
    def init(self):
        self.cache_encapes = {}
    def cache_get(self, key, ns):
        if key is None:
            return None
        key = (ns, key)
        return dz.get(self.cache_encapes, key, None)
    def cache_set(self, key, ns, encape):
        if key is None:
            return
        key = (ns, key)
        self.cache_encapes[key] = encape
    @staticmethod
    def get_encape(key, unit):
        if Confs.is_conf(key):
            ep,_,find = unit.get_encape(key, unit)
            assert find
            return ep
        return key
    def build(self, conf, unit):
        return None
    def deal(self, conf, unit):
        'encape, conf, conf_need_udpate'
        encape = self.build(conf,unit)
        return encape,conf,False
    def call(self, conf, unit):
        'encape, conf, conf_need_udpate'
        id,find=unit.conf_key(conf)
        ns = unit.ns
        encape = self.cache_get(id, ns)
        if encape is not None:
            return encape, conf, False
        encape,conf,upd = self.deal(conf,unit)
        self.cache_set(id, ns, encape)
        return encape,conf,upd

pass




from buildz.iocz.ioc import *
from buildz.iocz.ioc_deal import obj, deal
from buildz import xf, pyz, Base
class Test(Base):
    def init(self, id=0):
        self.id = id

pass
conf_obj = xf.loads(r"""
type=obj
source=buildz.iocz.ioc.test.Test
args=[
    {
        type=obj
        source = buildz.iocz.ioc.test.Test
        maps={id=123}
    }
]
""")
conf_init = xf.loads(r"""
type=deal
source=buildz.iocz.ioc_deal.obj.ObjectDeal
deals: obj
call=1
""")
def test():
    mg = Manager('.', 'type')
    mg.set_deal("deal", deal.DealDeal())
    print(mg)
    unit = mg.create()
    print(unit)
    #deal_obj = obj.ObjectDeal()
    #mg.set_deal('obj', deal_obj)
    unit.set_conf("test",conf_obj)
    unit.set_conf("deal_obj",conf_init)
    unit.add_build("deal_obj")
    encape, tag, find = mg.get_encape("test")
    print(f"encape: {encape}")
    it = encape()
    print(f"obj: {id(it), id(it.id),it.id.id}")
    it = encape()
    print(f"obj: {id(it), id(it.id),it.id.id}")

pyz.lc(locals(), test)


#

# type 1
confbd = ConfsBuilder()
confbd.add_build(EnvBuilder())
confsbd.add_build(DataBduiler())
confsbd.add_build(InitBuilder())
confsbd.add_build(VarBuilder())
confs = confsbd()

confbd = confsbd.build_conf()

conf = Conf()
conf.add_env()
conf.add_data()
conf.add_init()

confs.add_conf(conf)

# type2

mg = Manager()
mg.add("env", EnvBuilder())
mg.add("data", Data())

mg.data.add(id, Data())

mg.data.get()



#
from ..ioc.unit import Unit
from ... import dz,xf
from ..ioc.datas import Datas
'''
id|namespace|ns:
deal_id|deal_ns: 
envs.pub|envs.pri|envs.ns|envs: {
    
}
confs.pub|confs.pri|confs.ns|confs: [

]
confs.pub|confs.pri|confs.ns|confs: {

}
builds: [
    
]

or

[
    ...
]
'''
class ConfUnit(Unit):
    key_ns = ('id', 'namespace', 'ns')
    key_deal_ns = "deal_id,deal_ns,deal_namespace".split(",")
    key_pub = "confs.pub,confs".split(",")
    key_pri = "confs.pri,confs.prv".split(",")
    key_ns = "confs.ns,confs.namespace".split(",")
    key_builds = "builds,build".split(",")
    def init(self, conf, mg):
        if type(conf)==str:
            conf = xf.loads(conf)
        if dz.islist(conf):
            conf = {'confs': conf}
        ns = dz.get_one(conf, self.key_ns)
        deal_ns = dz.get_one(conf, self.key_deal_ns)
        super().init(ns, deal_ns)
        self.bind(mg)
        self.load(conf)
    def load_confs(self, confs, tag=None):
        rst = []
        if dz.isdict(confs):
            for k,v in confs.items():
                self.conf_key.fill(v, k)
                rst.append(v)
            confs = rst
        for item in confs:
            key,find = self.conf_key(item)
            if find:
                self.set_conf(key, item, tag)
    def load(self, conf):
        keys = [self.key_pub, self.key_ns, self.key_pri]
        tags = [Datas.Key.Pub, Datas.Key.Ns, Datas.Key.Pri]
        for key, tag in zip(keys, tags):
            tag_confs = dz.get_one(conf, key, [])
            self.load_confs(tag_confs, tag)
        builds = dz.get_one(conf, self.key_builds, [])
        for item in builds:
            self.add_build(item)

pass




from buildz import ioc

from buildz import xf, pyz, Base
class Test(Base):
    def init(self, id=0):
        self.id = id

pass
def test():
    conf = xf.loads(r"""
    [
        {
            id=test
            type=obj
            source=buildz.iocz.ioc.test.Test
            single=0
            args=[
                {
                    type=obj
                    source = buildz.iocz.ioc.test.Test
                    maps={id=123}
                }
            ]
        }
    ]
    """)
    mg = ioc.build()
    mg.add(conf)
    print(mg)
    it = mg.get("test")
    print(f"obj: {id(it), id(it.id),it.id.id}")
    it = mg.get("test")
    print(f"obj: {id(it), id(it.id),it.id.id}")

pyz.lc(locals(), test)